#!python3
# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Validation pipeline configuration data."""

import asyncio
import enum
import pathlib
import sys
import typing

import click

from ._version import acquire_version
from .console import report_failure, report_success
from .pipeline_config import (
    PipelineConfiguration,
    PipelineConfigurationPaths,
    do_path_check,
)
from .pipeline_config.exceptions import (
    ClientsDefinitionError,
    ConfigurationPathsError,
    DeploymentsDefinitionError,
    FrameDefinitionsError,
    PipelineConfigurationError,
    PipelineViewError,
    ReleaseStatesDefinitionError,
    SubscriptionsDefinitionError,
    SystemsDefinitionError,
    TenantsDefinitionError,
)
from .utilities import acquire_token


@enum.unique
class ExitState(enum.Enum):
    """Validate configuration script exit states."""

    UNKNOWN = 100
    MISSING_GITREF = 101
    GITREF_PARSE_FAILURE = 102
    BAD_CONFIGURATION_PATHS = 103


@click.command()
@click.version_option(acquire_version())
@click.argument(
    "client_path",
    type=click.Path(dir_okay=True, file_okay=False, path_type=pathlib.Path),
)
@click.argument(
    "system_path",
    type=click.Path(dir_okay=True, file_okay=False, path_type=pathlib.Path),
)
@click.argument(
    "password_file",
    type=click.File(mode="r"),
)
@click.option(
    "--check-paths",
    default=False,
    help="Check paths in configuration for file or directory existence.",
    is_flag=True,
)
@click.option(
    "--disable-vaults",
    default=False,
    help="""Disable encrypted file validation (local developer use only).

Disabling the validation still checks for the presence of the correct file,
but does not attempt to decrypt the contents for further validation in order
to avoid having to specify the decryption password.
""",
    is_flag=True,
)
def _main(
    client_path: pathlib.Path,
    system_path: pathlib.Path,
    password_file: typing.IO,
    check_paths: bool,
    disable_vaults: bool,
) -> None:
    """
    Validate pipeline configuration files.

    Exits non-zero if validation fails. Without the ``--check-paths`` option
    the utility loads the configuration to check for self-consistency of the
    data. In this case the release state does not need to be known.

    With the ``--check-paths`` option, the utility iterates over the release
    states configured for the client to check that all arm template and arm
    templates parameter files are in the expected locations unless the
    ``--git-ref`` option is also specified. In this case the check path tests
    only apply to the release state implied by the git ref.

    If ``--disable-sp`` is NOT specified the client service principal vault
    decryption password must be piped via stdin using an ``echo`` or ``cat``
    command.

    CLIENT_PATH:  The path to the directory where client files are located.
    SYSTEM_PATH:  The path to the directory where system files are located.
    PASSWORD_FILE:  The path to a file where the service principal decryption
                    password is stored, or "-" for stdin.
    """
    try:
        client_config = client_path / "configuration"
        system_config = system_path / "configuration"
        configuration_paths = PipelineConfigurationPaths.from_paths(
            client_config, system_config
        )

        decrypt_token = None
        if not disable_vaults:
            decrypt_token = acquire_token(password_file)

        pipeline_configuration = PipelineConfiguration.from_files(
            configuration_paths, decrypt_token
        )

        if check_paths:
            asyncio.run(do_path_check(pipeline_configuration, system_path))

        report_success("pipeline configuration validated")
    except FileNotFoundError as e:
        report_failure(str(e))
        sys.exit(ExitState.BAD_CONFIGURATION_PATHS.value)
    except ConfigurationPathsError as e:
        report_failure(str(e))
        sys.exit(ExitState.BAD_CONFIGURATION_PATHS.value)
    except PipelineConfigurationError as e:
        report_failure(str(e))
        sys.exit(ExitState.MISSING_GITREF.value)
    except (
        PipelineConfigurationError,
        ClientsDefinitionError,
        DeploymentsDefinitionError,
        FrameDefinitionsError,
        ReleaseStatesDefinitionError,
        SubscriptionsDefinitionError,
        SystemsDefinitionError,
        TenantsDefinitionError,
        PipelineViewError,
    ) as e:
        report_failure("Configuration schema error, {0}".format(e))
        sys.exit(ExitState.BAD_CONFIGURATION_PATHS.value)
    except Exception as e:
        report_failure("unknown error, {0}".format(str(e)))
        sys.exit(ExitState.UNKNOWN.value)


def flit_entry() -> None:
    """Flit script entry function for ``validate-configuration`` utility."""
    _main()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Utility related exceptions."""


class AnsibleVaultError(Exception):
    """Problem completing Ansible Vault operation."""


class CommandError(Exception):
    """Problem completing an external command run."""



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy

from foodx_devops_tools.deploy_me._deployment import _construct_deployment_name


def test_clean(mock_deploystep_context):
    mock_data = copy.deepcopy(mock_deploystep_context["deployment_data"])
    mock_data.context.pipeline_id = "12345678.24"

    result = _construct_deployment_name(mock_data, "stepname")

    assert result == "app-name_stepname_12345678.24"


def test_plus(mock_deploystep_context):
    mock_data = copy.deepcopy(mock_deploystep_context["deployment_data"])
    mock_data.context.pipeline_id = "12345678+24"

    result = _construct_deployment_name(mock_data, "stepname")

    assert result == "app-name_stepname_12345678-24"


def test_underscore(mock_deploystep_context):
    mock_data = copy.deepcopy(mock_deploystep_context["deployment_data"])
    mock_data.context.pipeline_id = "12345678_24"

    result = _construct_deployment_name(mock_data, "stepname")

    assert result == "app-name_stepname_12345678-24"



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
import pathlib

import pytest

from foodx_devops_tools.deploy_me._deployment import (
    ApplicationDeploymentDefinition,
    DeploymentState,
    _construct_arm_paths,
    _construct_fqdn,
    _construct_resource_group_name,
    assess_results,
)
from foodx_devops_tools.pipeline_config.frames import DeploymentMode


class TestConstructFqdn:
    def test_structured_subscription(self):
        result = _construct_fqdn(
            "api", "some.where", "client", "system_client_stub"
        )

        assert result == "api.stub.client.some.where"

    def test_unstructured_subscription(self):
        result = _construct_fqdn(
            "api", "some.where", "client", "system_client_stub"
        )

        assert result == "api.stub.client.some.where"


class TestConstructResourceGroupName:
    def test_clean(self):
        result = _construct_resource_group_name("app", "frame", "client")

        assert result == "app-frame-client"


class TestAssessResults:
    @pytest.mark.asyncio
    async def test_success(self):
        mock_results = [
            DeploymentState(code=DeploymentState.ResultType.success),
            DeploymentState(code=DeploymentState.ResultType.success),
        ]

        result = await assess_results(mock_results)

        assert result.code == DeploymentState.ResultType.success

    @pytest.mark.asyncio
    async def test_fail(self):
        mock_results = [
            DeploymentState(code=DeploymentState.ResultType.success),
            DeploymentState(code=DeploymentState.ResultType.failed),
        ]

        result = await assess_results(mock_results)

        assert result.code == DeploymentState.ResultType.failed


class TestConstructArmPaths:
    def test_none_arm_file_puff_file(self):
        mock_app_name = "this_app"
        mock_puff_parameter_name = "some.puff.file.json"
        mock_frame_folder = pathlib.Path("some/path")
        mock_step = ApplicationDeploymentDefinition(
            mode=DeploymentMode.incremental,
            name="somename",
            arm_file=None,
            puff_file=None,
        )

        result_template, result_puff, result_parameters = _construct_arm_paths(
            mock_step,
            mock_puff_parameter_name,
            mock_app_name,
            mock_frame_folder,
        )

        assert result_template == pathlib.Path("some/path/this_app.json")
        assert result_puff == pathlib.Path("some/path/this_app.yml")

    def test_parameter_file(self):
        mock_app_name = "this_app"
        mock_puff_parameter_name = "some.puff.file.json"
        mock_frame_folder = pathlib.Path("some/path")
        mock_step = ApplicationDeploymentDefinition(
            mode=DeploymentMode.incremental,
            name="somename",
        )

        result_template, result_puff, result_parameters = _construct_arm_paths(
            mock_step,
            mock_puff_parameter_name,
            mock_app_name,
            mock_frame_folder,
        )

        assert result_parameters == pathlib.Path(
            "some/path/some.puff.file.json"
        )

    def test_arm_file_none_puff_file(self):
        mock_app_name = "this_app"
        mock_puff_parameter_name = "some.puff.file.json"
        mock_frame_folder = pathlib.Path("some/path")
        mock_step = ApplicationDeploymentDefinition(
            mode=DeploymentMode.incremental,
            name="somename",
            arm_file=pathlib.Path("arm_file.json"),
            puff_file=None,
        )

        result_template, result_puff, result_parameters = _construct_arm_paths(
            mock_step,
            mock_puff_parameter_name,
            mock_app_name,
            mock_frame_folder,
        )

        assert result_template == pathlib.Path("some/path/arm_file.json")
        assert result_puff == pathlib.Path("some/path/arm_file.yml")

    def test_independent_arm_file_puff_file(self):
        mock_app_name = "this_app"
        mock_puff_parameter_name = "some.puff.file.json"
        mock_frame_folder = pathlib.Path("some/path")
        mock_step = ApplicationDeploymentDefinition(
            mode=DeploymentMode.incremental,
            name="somename",
            arm_file=pathlib.Path("arm_file.json"),
            puff_file=pathlib.Path("puff_file.yml"),
        )

        result_template, result_puff, result_parameters = _construct_arm_paths(
            mock_step,
            mock_puff_parameter_name,
            mock_app_name,
            mock_frame_folder,
        )

        assert result_template == pathlib.Path("some/path/arm_file.json")
        assert result_puff == pathlib.Path("some/path/puff_file.yml")



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Subscriptions deployment configuration I/O."""

import pathlib
import typing

import pydantic

from ._exceptions import DeploymentsDefinitionError
from ._loader import load_configuration

ENTITY_NAME = "deployments"


class DeploymentLocations(pydantic.BaseModel):
    """Define primary and/or secondary locations for deployment."""

    primary: str
    secondary: typing.Optional[str]


class DeploymentSubscriptionReference(pydantic.BaseModel):
    """A subscription reference in a deployment definition."""

    locations: typing.List[DeploymentLocations]


class SingularDeployment(pydantic.BaseModel):
    """Definition of a singular deployment."""

    subscriptions: typing.Dict[str, DeploymentSubscriptionReference]


T = typing.TypeVar("T", bound="DeploymentsDefinition")

ValueType = typing.Dict[str, SingularDeployment]


class DeploymentsDefinition(pydantic.BaseModel):
    """Definition of deployments."""

    deployments: ValueType

    @pydantic.validator(ENTITY_NAME)
    def check_deployments(
        cls: pydantic.BaseModel, value: ValueType
    ) -> ValueType:
        """Validate ``deployments`` field."""
        if not value:
            raise ValueError("Empty deployment names prohibited")

        return value


def load_deployments(file_path: pathlib.Path) -> DeploymentsDefinition:
    """
    Load client definitions from file.

    Args:
        file_path: Path to client definitions file.

    Returns:
        Deployment definitions.
    Raises:
        DeploymentsDefinitionError: If an error occurs loading the file.
    """
    result = load_configuration(
        file_path,
        DeploymentsDefinition,
        DeploymentsDefinitionError,
        ENTITY_NAME,
    )

    return typing.cast(DeploymentsDefinition, result)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Release flow package internal exceptions."""


class ReleaseIdentityError(Exception):
    """Problem parsing release id from git reference."""


class ReleaseStateError(Exception):
    """Problem parsing release state from git reference."""



# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from foodx_devops_tools.release_flow_entry import release_flow
from tests.ci.support.click_runner import click_runner  # noqa: F401


class TestNpmSubcommand:
    MOCK_PACKAGE_JSON_CONTENT = """{
  "author": "FoodX",
  "description": "some package",
  "keywords": [
    "foodx"
  ],
  "license": "SEE LICENSE IN LICENSE",
  "name": "@foodx/some-package",
  "version": "0.0.0+local"
}
"""

    def test_npm_id_nonrelease(self, click_runner, mocker):
        mock_input = [
            "npm",
            "id",
            "package.json",
            "refs/tags/3.14.159-alpha.13",
            "abc123def",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        result = click_runner.invoke(release_flow, mock_input)

        assert result.exit_code == 0
        assert result.output == "3.1.4-post.26.abc123d"

    def test_npm_id_release(self, click_runner, mocker):
        mock_input = [
            "npm",
            "id",
            "package.json",
            "refs/tags/3.14.159",
            "abc123def",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        result = click_runner.invoke(release_flow, mock_input)

        assert result.exit_code == 0
        assert result.output == "3.14.159"

    def test_npm_package_nonrelease(self, click_runner, mocker):
        mock_input = [
            "npm",
            "package",
            "package.json",
            "refs/tags/3.14.159-alpha.13",
            "abc123def",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        with click_runner.isolated_filesystem():
            with open("package.json", mode="w") as f:
                f.write(self.MOCK_PACKAGE_JSON_CONTENT)
            result = click_runner.invoke(release_flow, mock_input)

        assert result.exit_code == 0
        assert result.output == "foodx-some-package-3.1.4-post.26.abc123d.tgz"

    def test_npm_package_release(self, click_runner, mocker):
        mock_input = [
            "npm",
            "package",
            "package.json",
            "refs/tags/3.14.159",
            "abc123def",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        with click_runner.isolated_filesystem():
            with open("package.json", mode="w") as f:
                f.write(self.MOCK_PACKAGE_JSON_CONTENT)
            result = click_runner.invoke(release_flow, mock_input)

        assert result.exit_code == 0
        assert result.output == "foodx-some-package-3.14.159.tgz"

    def test_main_branch(self, click_runner, mocker):
        mock_arguments = [
            "npm",
            "package",
            "package.json",
            "refs/heads/main",
            "123abc",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow.npm_ci.apply_package_release_id",
            return_value="@some-group/this-package",
        )
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        with click_runner.isolated_filesystem():
            with open("package.json", mode="w") as f:
                f.write(self.MOCK_PACKAGE_JSON_CONTENT)
            result = click_runner.invoke(release_flow, mock_arguments)

        assert result.exit_code == 0
        assert (
            result.output == "some-group-this-package-3.1.4-post.26.123abc.tgz"
        )

    def test_release_tag(self, click_runner, mocker):
        mock_arguments = [
            "npm",
            "package",
            "package.json",
            "refs/tags/3.14.159",
            "123abc",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow.npm_ci.apply_package_release_id",
            return_value="@some-group/this-package",
        )
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        with click_runner.isolated_filesystem():
            with open("package.json", mode="w") as f:
                f.write(self.MOCK_PACKAGE_JSON_CONTENT)
            result = click_runner.invoke(release_flow, mock_arguments)

        assert result.exit_code == 0
        assert result.output == "some-group-this-package-3.14.159.tgz"

    def test_dryrun_tag(self, click_runner, mocker):
        mock_arguments = [
            "npm",
            "package",
            "package.json",
            "refs/tags/3.14.159-dryrun45",
            "123abc",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow.npm_ci.apply_package_release_id",
            return_value="@some-group/this-package",
        )
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        with click_runner.isolated_filesystem():
            with open("package.json", mode="w") as f:
                f.write(self.MOCK_PACKAGE_JSON_CONTENT)
            result = click_runner.invoke(release_flow, mock_arguments)

        assert result.exit_code == 0
        assert result.output == "some-group-this-package-3.14.159-dryrun45.tgz"



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Python clone of node.js puff utility."""

import asyncio
import enum
import logging
import pathlib
import sys

import click

from foodx_devops_tools._version import acquire_version

from ._declarations import (
    DEFAULT_CONSOLE_LOGGING_ENABLED,
    DEFAULT_FILE_LOGGING_DISABLED,
    DEFAULT_LOG_LEVEL,
    VALID_LOG_LEVELS,
)
from ._logging import LoggingState
from .puff import PuffError, run_puff

log = logging.getLogger(__name__)

DEFAULT_LOG_FILE = pathlib.Path("puff_run.log")


@enum.unique
class ExitState(enum.Enum):
    """Numeric script exit states."""

    UNKNOWN = 100
    PUFF_FAILED = 101


@click.command()
@click.version_option(version=acquire_version())
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=True, dir_okay=True),
)
@click.option(
    "--delete",
    "-d",
    default=False,
    help="Delete files that were generated.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--log-enable-console",
    "enable_console_log",
    default=DEFAULT_CONSOLE_LOGGING_ENABLED,
    help="Log to console.",
    is_flag=True,
)
@click.option(
    "--log-disable-file",
    "disable_file_log",
    default=DEFAULT_FILE_LOGGING_DISABLED,
    help="Disable file logging.",
    is_flag=True,
)
@click.option(
    "--log-level",
    "log_level",
    default=DEFAULT_LOG_LEVEL,
    help="Select logging level to apply to all enabled log sinks.",
    show_default=True,
    type=click.Choice(VALID_LOG_LEVELS, case_sensitive=False),
)
@click.option(
    "--pretty",
    default=False,
    help="Create nicely formatted JSON for humans.",
    is_flag=True,
    show_default=True,
)
def _main(
    path: str,
    delete: bool,
    disable_file_log: bool,
    enable_console_log: bool,
    log_level: str,
    pretty: bool,
) -> None:
    """
    Create or delete ARM template parameter files from puff YAML configuration.

    If a directory is specified, all the yaml files in the directory (
    recursively) are assumed to be puff parameter files. If a file is
    specified then just the file is assumed to be a puff parameter file and
    in this case the json files are generated to the parameter files parent
    directory.

    PATH    Directory or file path for finding yml files to generate from.
    """
    try:
        # currently no need to change logging configuration at run time,
        # so no need to preserve the object.
        LoggingState(
            disable_file_logging=disable_file_log,
            enable_console_logging=enable_console_log,
            log_level_text=log_level,
            default_log_file=DEFAULT_LOG_FILE,
        )

        asyncio.run(run_puff(pathlib.Path(path), delete, pretty))
    except PuffError as e:
        click.echo(str(e), err=True)
        sys.exit(ExitState.PUFF_FAILED.value)
    except asyncio.exceptions.CancelledError:
        log.warning("Exiting due to cancellation")
        # async CancelledError exceptions should usually be re-raised
        # https://docs.python.org/3.8/library/asyncio-exceptions.html#asyncio.CancelledError
        raise
    except Exception as e:
        click.echo(
            "Puff failed with unexpected error (see log for "
            "details), {0}".format(str(e)),
            err=True,
        )
        log.exception("Unexpected error, {0}".format(str(e)))
        sys.exit(ExitState.UNKNOWN.value)


def entrypoint() -> None:
    """Flit entry point."""
    _main()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Run resource group tests against an actual Azure Cloud subscription."""

import logging

import pytest

from foodx_devops_tools.azure.cloud.resource_group import (
    check_exists as check_resource_group_exists,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    create as create_resource_group,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    delete as delete_resource_group,
)

from .support import (  # noqa: F401
    clean_teardown,
    prexisting_resource_group,
    unique_group_name,
)

log = logging.getLogger(__name__)


class TestCheckResourceGroupExists:
    @pytest.mark.asyncio
    async def test_clean(
        self, prexisting_resource_group, unique_group_name, user_subscription
    ):
        expected_name = unique_group_name
        assert not await check_resource_group_exists(
            expected_name, user_subscription
        )
        async with prexisting_resource_group(expected_name):
            result = await check_resource_group_exists(
                expected_name, user_subscription
            )

            assert result["name"] == expected_name


class TestCreateResourceGroup:
    @pytest.mark.asyncio
    async def test_clean(
        self, clean_teardown, unique_group_name, user_subscription
    ):
        expected_name = unique_group_name
        async with clean_teardown(expected_name):
            await create_resource_group(
                expected_name, "West US 2", user_subscription
            )

            assert await check_resource_group_exists(
                expected_name, user_subscription
            )


class TestDeleteResourceGroup:
    @pytest.mark.asyncio
    async def test_clean(
        self, prexisting_resource_group, unique_group_name, user_subscription
    ):
        expected_name = unique_group_name
        async with prexisting_resource_group(expected_name):
            await delete_resource_group(expected_name, user_subscription)

            assert not await check_resource_group_exists(
                expected_name, user_subscription
            )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import logging
import pathlib
import typing

from foodx_devops_tools.puff_utility import ExitState, PuffError, _main
from tests.ci.support.click_runner import (  # noqa: F401
    click_runner,
    isolated_filesystem,
)


@contextlib.contextmanager
def multifile_fs(
    expected_dir: str, this_runner
) -> typing.Generator[pathlib.Path, None, None]:
    with isolated_filesystem(expected_dir, this_runner) as temp_dir:
        with (temp_dir / "y1.yml").open(mode="w") as f:
            f.write(
                """---
name: some-y1
environments: {}
"""
            )

        with (temp_dir / "y2.yml").open(mode="w") as f:
            f.write(
                """---
name: some-y2
environments: {}
"""
            )
        yield temp_dir


@contextlib.contextmanager
def singlefile_fs(
    expected_dir: str, this_runner
) -> typing.Generator[pathlib.Path, None, None]:
    with isolated_filesystem(expected_dir, this_runner) as temp_dir:
        filename = temp_dir / "y1.yml"
        with filename.open(mode="w") as f:
            f.write(
                """---
name: some-y1
environments: {}
"""
            )
        yield filename


class TestMain:
    def test_help(self, caplog, click_runner, mocker):
        expected_dir = "some_path"

        mocker.patch(
            "foodx_devops_tools.puff.run.load_puffignore",
            return_value=list(),
        )

        with singlefile_fs(expected_dir, click_runner) as puff_file:
            result = click_runner.invoke(_main, ["--help"])

            assert result.exit_code == 0
            assert (
                "Create or delete ARM template parameter files from puff "
                "YAML configuration." in result.output
            )

    def test_single_file(self, caplog, click_runner, mocker):
        expected_dir = "some_path"
        caplog.set_level(logging.WARN)

        mocker.patch(
            "foodx_devops_tools.puff.run.load_puffignore",
            return_value=list(),
        )

        with singlefile_fs(expected_dir, click_runner) as puff_file:
            result = click_runner.invoke(_main, [str(puff_file)])

            message = "No puff parameter files found in directory"
            assert message not in caplog.text
            assert result.exit_code == 0

    def test_multifile(self, caplog, click_runner, mocker):
        expected_dir = "some_path"
        caplog.set_level(logging.WARN)

        mocker.patch(
            "foodx_devops_tools.puff.run.load_puffignore",
            return_value=list(),
        )

        with multifile_fs(expected_dir, click_runner) as temp_dir:
            result = click_runner.invoke(_main, [str(temp_dir)])

            assert result.exit_code == 0
            message = "No puff parameter files found in directory"
            assert message not in caplog.text

    def test_delete(self, caplog, click_runner, mocker):
        expected_dir = "some_path"

        mocker.patch(
            "foodx_devops_tools.puff.run.load_puffignore",
            return_value=list(),
        )

        for arg_form in ["-d", "--delete"]:
            with multifile_fs(expected_dir, click_runner) as temp_dir:
                result = click_runner.invoke(_main, [str(temp_dir), arg_form])

            assert result.exit_code == 0
            message = "No puff parameter files found in directory"
            assert message not in caplog.text

    def test_puff_error_exits_dirty(self, click_runner, mocker):
        expected_dir = "some_path"
        error_message = "some puff error"

        mocker.patch(
            "foodx_devops_tools.puff.run.load_puffignore",
            side_effect=PuffError(error_message),
        )

        with isolated_filesystem(expected_dir, click_runner) as temp_dir:
            result = click_runner.invoke(_main, [str(temp_dir)])

            assert result.exit_code == ExitState.PUFF_FAILED.value
            assert error_message in result.output

    def test_unknown_error_exits_dirty(self, click_runner, mocker):
        expected_dir = "some_path"
        error_message = "some unknown  error"

        mocker.patch(
            "foodx_devops_tools.puff.run.load_puffignore",
            side_effect=RuntimeError(error_message),
        )

        with isolated_filesystem(expected_dir, click_runner) as temp_dir:
            result = click_runner.invoke(_main, [str(temp_dir)])

            assert result.exit_code == ExitState.UNKNOWN.value
            assert "Puff failed with unexpected error" in result.output
            assert error_message in result.output



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import pathlib
import typing

import aiofiles  # type: ignore

from foodx_devops_tools.puff import run_puff

from ._structure import StructuredPathCollection
from .frames import FramesTriggersDefinition
from .pipeline import PipelineConfiguration
from .puff_map import PuffMap

FileExistsCollection = typing.Dict[pathlib.Path, bool]


def _collate_results(
    results: typing.List[FileExistsCollection],
) -> FileExistsCollection:
    collated_results: FileExistsCollection = dict()
    for this_result in results:
        collated_results.update(this_result)

    return collated_results


async def _file_exists(file: pathlib.Path) -> bool:
    """Test for file existence asynchronously."""
    try:
        async with aiofiles.open(str(file), mode="r"):
            pass

        exists = True
    except FileNotFoundError:
        exists = False

    return exists


async def _check_paths(paths: typing.Set[pathlib.Path]) -> FileExistsCollection:
    results = await asyncio.gather(*[_file_exists(x) for x in paths])

    collection: FileExistsCollection = dict()
    for file, state in zip(paths, results):
        collection[file] = state

    return collection


async def _check_arm_files(
    frames: FramesTriggersDefinition,
) -> FileExistsCollection:
    return await _check_paths(set(frames.arm_file_paths().values()))


async def _check_puff_maps(
    frame_folders: StructuredPathCollection, puff_map: PuffMap
) -> FileExistsCollection:
    return await _check_paths(
        set(puff_map.arm_template_parameter_file_paths(frame_folders).values())
    )


async def do_path_check(
    pipeline_configuration: PipelineConfiguration, system_path: pathlib.Path
) -> None:
    """Check that paths in configuration actually exist."""
    # check that all the generated ARM template parameter files and ARM
    # template files are where they are expected to be for the current
    # *client* configuration.

    await run_puff(system_path, False, False, disable_ascii_art=True)

    this_futures = await asyncio.gather(
        _check_arm_files(pipeline_configuration.frames),
        _check_puff_maps(
            pipeline_configuration.frames.frame_folders(),
            pipeline_configuration.puff_map,
        ),
        return_exceptions=False,
    )

    done_items = [x for x in this_futures if x]
    collated_results = _collate_results(done_items)
    missing_files = [str(x) for x, y in collated_results.items() if not y]
    if missing_files:
        raise FileNotFoundError(
            "files missing from deployment, {0}".format(str(missing_files))
        )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""
Azure cloud utilities.

Exports:

* AzureSubscriptionConfiguration class
* deploy_resource_group function
"""

from .auth import AzureCredentials, login_service_principal  # noqa: F401
from .model import AzureSubscriptionConfiguration  # noqa: F401
from .resource_group import deploy as deploy_resource_group  # noqa: F401



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.deployment import DeploymentTuple, DeploymentTupleError


def test_clean():
    under_test = DeploymentTuple(
        client="this_client", release_state="this_release", system="this_system"
    )


def test_str():
    under_test = DeploymentTuple(
        client="this_client", release_state="this_release", system="this_system"
    )

    assert str(under_test) == "this_system-this_client-this_release"


def test_parse_clean():
    expected_text = "this_system-this_client-this_release"
    under_test = DeploymentTuple.parse(expected_text)

    assert str(under_test) == "this_system-this_client-this_release"


def test_parse_raises():
    expected_text = "this_system-NOTVALID??-this_release"
    with pytest.raises(DeploymentTupleError):
        under_test = DeploymentTuple.parse(expected_text)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Manage pipeline configuration files collectively."""

import logging
import re
import typing

import pydantic

from ._exceptions import PipelineConfigurationError
from ._paths import PipelineConfigurationPaths
from .clients import ValueType as ClientsData
from .clients import load_clients
from .deployments import ValueType as DeploymentsData
from .deployments import load_deployments
from .frames import ValueType as FramesData
from .frames import load_frames
from .puff_map import PuffMap, load_puff_map
from .release_states import ValueType as ReleaseStatesData
from .release_states import load_release_states
from .service_principals import ValueType as ServicePrincipalsData
from .service_principals import load_service_principals
from .static_secrets import ValueType as StaticSecrets
from .static_secrets import load_static_secrets
from .subscriptions import ValueType as SubscriptionsData
from .subscriptions import load_subscriptions
from .systems import ValueType as SystemsData
from .systems import load_systems
from .tenants import ValueType as TenantsData
from .tenants import load_tenants

log = logging.getLogger(__name__)

DEPLOYMENT_NAME_REGEX = (
    r"^(?P<system>[a-z0-9]+)"
    r"[_-]"
    r"(?P<client>[a-z0-9]+)"
    r"[_-]"
    r"(?P<release_state>[a-z0-9]+)$"
)


T = typing.TypeVar("T", bound="PipelineConfiguration")


class PipelineConfiguration(pydantic.BaseModel):
    """Pipeline configuration data."""

    clients: ClientsData
    release_states: ReleaseStatesData
    deployments: DeploymentsData
    frames: FramesData
    puff_map: PuffMap
    service_principals: typing.Optional[ServicePrincipalsData]
    static_secrets: typing.Optional[StaticSecrets]
    subscriptions: SubscriptionsData
    systems: SystemsData
    tenants: TenantsData

    decrypt_token: typing.Optional[str] = None

    @classmethod
    def from_files(
        cls: typing.Type[T],
        paths: PipelineConfigurationPaths,
        decrypt_token: typing.Optional[str],
    ) -> T:
        """
        Load pipeline configuration from collection of files.

        Args:
            paths: Paths to pipeline configuration files.

        Returns:
            Instantiated ``PipelineConfiguration`` object.
        """
        client_config = load_clients(paths.clients)
        release_state_config = load_release_states(paths.release_states)
        deployment_config = load_deployments(paths.deployments)
        frames_config = load_frames(paths.frames)
        puff_map_config = load_puff_map(paths.puff_map)
        subscription_config = load_subscriptions(paths.subscriptions)
        system_config = load_systems(paths.systems)
        tenant_config = load_tenants(paths.tenants)

        service_principals_config = None
        static_secrets_config = None
        if decrypt_token:
            service_principals_config = load_service_principals(
                paths.service_principals, decrypt_token
            )
            static_secrets_config = load_static_secrets(
                paths.static_secrets, decrypt_token
            )
        else:
            if not paths.service_principals.is_file():
                raise FileNotFoundError(
                    "Missing service principals vault "
                    "file, {0}".format(paths.service_principals)
                )
            missing_files = {
                str(x) for x in paths.static_secrets if not x.is_file()
            }
            if any(missing_files):
                raise FileNotFoundError(
                    "Missing static secrets files, {0}".format(
                        str(missing_files)
                    )
                )

        kwargs = {
            "clients": client_config.clients,
            "release_states": release_state_config.release_states,
            "deployments": deployment_config.deployments,
            "frames": frames_config.frames,
            "puff_map": puff_map_config.puff_map,
            "subscriptions": subscription_config.subscriptions,
            "systems": system_config.systems,
            "tenants": tenant_config.tenants,
            "decrypt_token": decrypt_token,
        }
        if service_principals_config:
            kwargs[
                "service_principals"
            ] = service_principals_config.service_principals
        if static_secrets_config:
            kwargs["static_secrets"] = static_secrets_config.static_secrets

        new_object = cls(**kwargs)

        return new_object

    @pydantic.root_validator()
    def check_clients(cls: pydantic.BaseModel, loaded_data: dict) -> dict:
        """Cross-check clients data with other fields."""
        for name, data in loaded_data["clients"].items():
            if not all(
                [
                    x in loaded_data["release_states"]
                    for x in data.release_states
                ]
            ):
                message = "Bad release state in client, {0}".format(name)
                log.error(
                    "{0}, {1}, {2}".format(
                        message,
                        str(loaded_data["release_states"]),
                        str(data.release_states),
                    )
                )
                raise PipelineConfigurationError(message)

            if data.system not in loaded_data["systems"]:
                message = "Bad system in client, {0}".format(name)
                log.error(
                    "{0}, {1}, {2}".format(
                        message,
                        str(loaded_data["release_states"]),
                        str(data.release_states),
                    )
                )
                raise PipelineConfigurationError(message)
        return loaded_data

    @pydantic.root_validator()
    def check_deployments(cls: pydantic.BaseModel, loaded_data: dict) -> dict:
        """Cross-check loaded deployment data."""
        deployment_names = set(loaded_data["deployments"].keys())
        this_re = re.compile(DEPLOYMENT_NAME_REGEX)
        for this_name in deployment_names:
            result = this_re.match(this_name)

            if not result:
                message = "Bad deployment tuple, {0}".format(this_name)
                log.error("{0}, {1}".format(message, DEPLOYMENT_NAME_REGEX))
                raise PipelineConfigurationError(message)

            this_client = result.group("client")
            if this_client not in loaded_data["clients"]:
                message = "Bad client in deployment tuple, {0}".format(
                    this_client
                )
                log.error(
                    "{0}, {1}, {2}".format(
                        message, this_name, str(loaded_data["clients"])
                    )
                )
                raise PipelineConfigurationError(message)

            this_release_state = result.group("release_state")
            if this_release_state not in loaded_data["release_states"]:
                message = "Bad release state in deployment tuple, {0}".format(
                    this_release_state
                )
                log.error(
                    "{0}, {1}, {2}".format(
                        message, this_name, str(loaded_data["release_states"])
                    )
                )
                raise PipelineConfigurationError(message)

            this_system = result.group("system")
            if this_system not in loaded_data["systems"]:
                message = "Bad system in deployment tuple, {0}".format(
                    this_system
                )
                log.error(
                    "{0}, {1}, {2}".format(
                        message, this_name, str(loaded_data["systems"])
                    )
                )
                raise PipelineConfigurationError(message)

            this_subscriptions = loaded_data["deployments"][
                this_name
            ].subscriptions
            for subscription_name in this_subscriptions.keys():
                if subscription_name not in loaded_data["subscriptions"]:
                    message = "Bad subscription in deployment, {0}".format(
                        this_name
                    )
                    log.error(
                        "{0}, {1}, {2}, {3}".format(
                            message,
                            this_name,
                            str(this_subscriptions.keys()),
                            str(loaded_data["subscriptions"]),
                        )
                    )
                    raise PipelineConfigurationError(message)
        return loaded_data

    @pydantic.root_validator()
    def check_service_principals(
        cls: pydantic.BaseModel, loaded_data: dict
    ) -> dict:
        """Validate service principal data."""
        if loaded_data["service_principals"]:
            bad_subscriptions = [
                "{0}".format(name)
                for name in loaded_data["service_principals"].keys()
                if name not in loaded_data["subscriptions"].keys()
            ]
            if any(bad_subscriptions):
                message = "Bad subscription(s) in service_principals"
                log.error("{0}, {1}".format(message, str(bad_subscriptions)))
                raise PipelineConfigurationError(message)
        return loaded_data

    @pydantic.root_validator()
    def check_static_secrets(
        cls: pydantic.BaseModel, loaded_data: dict
    ) -> dict:
        """Validate static secret data."""
        if loaded_data["static_secrets"]:
            bad_subscriptions = [
                "{0}".format(name)
                for name in loaded_data["static_secrets"].keys()
                if name not in loaded_data["subscriptions"].keys()
            ]
            if any(bad_subscriptions):
                message = "Bad subscription(s) in static_secrets"
                log.error("{0}, {1}".format(message, str(bad_subscriptions)))
                raise PipelineConfigurationError(message)
        return loaded_data

    @pydantic.root_validator()
    def check_subscriptions(cls: pydantic.BaseModel, loaded_data: dict) -> dict:
        """Validate subscriptions data."""
        bad_tenants = [
            "{0}.{1}".format(name, data.tenant)
            for name, data in loaded_data["subscriptions"].items()
            if data.tenant not in loaded_data["tenants"]
        ]
        if any(bad_tenants):
            message = "Bad tenant(s) in subscription"
            log.error(
                "{0}, {1}, {2}".format(
                    message, str(bad_tenants), str(loaded_data["tenants"])
                )
            )
            raise PipelineConfigurationError(message)
        return loaded_data

    @pydantic.root_validator()
    def check_frames_puff_map(
        cls: pydantic.BaseModel, loaded_data: dict
    ) -> dict:
        """
        Cross-check frames, puff map data.

        * The identified frames must be the same.
        * The identified applications must be the same.
        * Any application deployment steps defined in a frame must be defined
          in the puff map.
        * Release states and subscriptions must be valid.
        """
        if set(loaded_data["frames"].frames.keys()) != set(
            loaded_data["puff_map"].frames.keys()
        ):
            message = "Frame definitions mismatch between frames and puff map"
            log.error(
                "{0}, {1}, {2}".format(
                    message,
                    str(set(loaded_data["frames"].frames.keys())),
                    str(set(loaded_data["puff_map"].frames.keys())),
                )
            )
            raise PipelineConfigurationError(message)
        for this_frame, frame_data in loaded_data["frames"].frames.items():
            if set(frame_data.applications.keys()) != set(
                loaded_data["puff_map"].frames[this_frame].applications.keys()
            ):
                message = (
                    "Application definitions mismatch between frames "
                    "and puff map"
                )
                log.error(
                    "{0}, {1}, {2}".format(
                        message,
                        str(set(frame_data.applications.keys())),
                        str(
                            set(
                                loaded_data["puff_map"]
                                .frames[this_frame]
                                .applications.keys()
                            )
                        ),
                    )
                )
                raise PipelineConfigurationError(message)
            for (
                this_application,
                application_data,
            ) in frame_data.applications.items():
                pm_app_data = (
                    loaded_data["puff_map"]
                    .frames[this_frame]
                    .applications[this_application]
                )
                if any(
                    [
                        x not in loaded_data["release_states"]
                        for x in pm_app_data.arm_parameters_files.keys()
                    ]
                ):
                    message = "Bad release state in puff map"
                    log.error(
                        "{0}, {1}, {2}".format(
                            message,
                            str(loaded_data["release_states"]),
                            str(pm_app_data.arm_parameters_files.keys()),
                        )
                    )
                    raise PipelineConfigurationError(message)
                for (
                    this_state,
                    state_data,
                ) in pm_app_data.arm_parameters_files.items():
                    if any(
                        [
                            x not in loaded_data["subscriptions"].keys()
                            for x in state_data.keys()
                        ]
                    ):
                        message = "Bad subscription in puff map"
                        log.error(
                            "{0}, {1}, {2}".format(
                                message,
                                str(loaded_data["subscriptions"].keys()),
                                str(state_data.keys()),
                            )
                        )
                        raise PipelineConfigurationError(message)
                    for subscription_data in state_data.values():
                        frame_step_names = {
                            x.name
                            for x in application_data
                            if hasattr(x, "name")
                        }
                        if frame_step_names != set(subscription_data.keys()):
                            message = (
                                "Application step name mismatch between "
                                "frames and puff map"
                            )
                            log.error(
                                "{0}, {1}, {2}".format(
                                    message,
                                    str(frame_step_names),
                                    str(set(subscription_data.keys())),
                                )
                            )
                            raise PipelineConfigurationError(message)
        return loaded_data



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import logging
import pathlib
import tempfile

import pytest

from foodx_devops_tools.pipeline_config import PipelineConfiguration
from foodx_devops_tools.pipeline_config._checks import (
    _file_exists,
    do_path_check,
)
from tests.ci.support.pipeline_config import MOCK_PATHS, MOCK_SECRET

log = logging.getLogger(__name__)


MOCK_SYSTEM_PATH = pathlib.Path("some/path")


class TestFileExists:
    @pytest.mark.asyncio
    async def test_true(self):
        with tempfile.TemporaryDirectory() as dir:
            this_dir = pathlib.Path(dir)
            this_file = this_dir / "exists"
            with this_file.open(mode="w") as f:
                f.write("")

            assert await _file_exists(this_file)

    @pytest.mark.asyncio
    async def test_false(self):
        with tempfile.TemporaryDirectory() as dir:
            this_dir = pathlib.Path(dir)
            this_file = this_dir / "notexists"

            assert not this_file.is_file()

            assert not await _file_exists(this_file)


class TestDoPathCheck:
    @pytest.mark.asyncio
    async def test_clean(
        self, mock_loads, mock_results, mock_async_method, mock_run_puff_check
    ):
        mock_async_method(
            "foodx_devops_tools.pipeline_config._checks._file_exists",
            return_value=True,
        )
        mock_loads(mock_results)
        mock_config = PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

        await do_path_check(mock_config, MOCK_SYSTEM_PATH)

    @pytest.mark.asyncio
    async def test_missing_file(
        self, mock_loads, mock_results, mock_async_method, mock_run_puff_check
    ):
        mock_async_method(
            "foodx_devops_tools.pipeline_config._checks._file_exists",
            side_effect=[True, False],
        )
        mock_loads(mock_results)
        mock_config = PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

        with pytest.raises(
            FileNotFoundError, match=r"files missing from " r"deployment"
        ):
            await do_path_check(mock_config, MOCK_SYSTEM_PATH)

    @pytest.mark.asyncio
    async def test_multiple_missing_files(
        self, mock_loads, mock_results, mock_async_method, mock_run_puff_check
    ):
        mock_async_method(
            "foodx_devops_tools.pipeline_config._checks" "._file_exists",
            side_effect=[False, False],
        )
        mock_loads(mock_results)
        mock_config = PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

        with pytest.raises(
            FileNotFoundError, match=r"files missing from " r"deployment"
        ):
            await do_path_check(mock_config, MOCK_SYSTEM_PATH)

    @pytest.mark.asyncio
    async def test_failed_check(
        self,
        mock_loads,
        mock_results,
        mock_async_method,
        mocker,
        mock_run_puff_check,
    ):
        mocker.patch(
            "foodx_devops_tools.pipeline_config._checks._check_arm_files",
            side_effect=RuntimeError(),
        )
        mock_async_method(
            "foodx_devops_tools.pipeline_config._checks._file_exists",
            return_value=True,
        )
        mock_loads(mock_results)
        mock_config = PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

        with pytest.raises(RuntimeError):
            await do_path_check(mock_config, MOCK_SYSTEM_PATH)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Context sensitive views of pipeline configuration data."""

import copy
import dataclasses
import logging
import pathlib
import typing

from foodx_devops_tools._to import StructuredTo
from foodx_devops_tools.azure.cloud import AzureCredentials
from foodx_devops_tools.patterns import SubscriptionData

from ..deployment import DeploymentTuple
from ._exceptions import PipelineViewError
from .pipeline import PipelineConfiguration, PuffMap

log = logging.getLogger(__name__)

Z = typing.TypeVar("Z", bound="IterationContext")


class IterationContext(typing.List[str]):
    """Record the iteration path."""

    def __str__(self: Z) -> str:
        """Convert the iteration path to a structured string."""
        return ".".join(self)


Y = typing.TypeVar("Y", bound="DeploymentContext")


class DeploymentContext:
    """Deployment context data expected to be applied to resource tags."""

    commit_sha: str
    pipeline_id: str
    release_id: str
    release_state: str

    # these data must be provided at deployment time because they are more
    # deeply context sensitive.
    __application_name: typing.Optional[str] = None
    __azure_subscription_name: typing.Optional[str] = None
    __azure_tenant_name: typing.Optional[str] = None
    __client: typing.Optional[str] = None
    __frame_name: typing.Optional[str] = None
    __system: typing.Optional[str] = None

    def __init__(
        self: Y,
        commit_sha: str,
        pipeline_id: str,
        release_id: str,
        release_state: str,
    ) -> None:
        """Construct ``DeploymentContext`` object."""
        self.commit_sha = commit_sha
        self.pipeline_id = pipeline_id
        self.release_id = release_id
        self.release_state = release_state

    @property
    def application_name(self: Y) -> str:
        """Get application name property."""
        if self.__application_name:
            return self.__application_name
        else:
            raise PipelineViewError(
                "application name not specified in deployment data"
            )

    @application_name.setter
    def application_name(self: Y, v: str) -> None:
        """Set application name property."""
        self.__application_name = v

    @property
    def azure_subscription_name(self: Y) -> str:
        """Get azure subscription name property."""
        if self.__azure_subscription_name:
            return self.__azure_subscription_name
        else:
            raise PipelineViewError(
                "azure subscription name not specified in deployment data"
            )

    @azure_subscription_name.setter
    def azure_subscription_name(self: Y, v: str) -> None:
        """Set azure subscription name property."""
        self.__azure_subscription_name = v

    @property
    def azure_tenant_name(self: Y) -> str:
        """Get azure tenant name property."""
        if self.__azure_tenant_name:
            return self.__azure_tenant_name
        else:
            raise PipelineViewError(
                "azure tenant name not specified in deployment data"
            )

    @azure_tenant_name.setter
    def azure_tenant_name(self: Y, v: str) -> None:
        """Set azure tenant name property."""
        self.__azure_tenant_name = v

    @property
    def client(self: Y) -> str:
        """Get client name property."""
        if self.__client:
            return self.__client
        else:
            raise PipelineViewError("client not specified in deployment data")

    @client.setter
    def client(self: Y, v: str) -> None:
        """Set client name property."""
        self.__client = v

    @property
    def frame_name(self: Y) -> str:
        """Get frame name property."""
        if self.__frame_name:
            return self.__frame_name
        else:
            raise PipelineViewError(
                "frame name not specified in deployment data"
            )

    @frame_name.setter
    def frame_name(self: Y, v: str) -> None:
        """Set frame name property."""
        self.__frame_name = v

    @property
    def resource_suffix(self: Y) -> str:
        """Get resource suffix property."""
        try:
            subscription_data = SubscriptionData.from_subscription_name(
                self.azure_subscription_name
            )

            return subscription_data.resource_suffix
        except PipelineViewError:
            # azure_subscription_name is undefined so just return an
            # identifiable response here.
            return "UNSPECIFIED"

    @property
    def system(self: Y) -> str:
        """Get system name property."""
        if self.__system:
            return self.__system
        else:
            raise PipelineViewError("system not specified in deployment data")

    @system.setter
    def system(self: Y, v: str) -> None:
        """Set system name property."""
        self.__system = v

    def as_dict(self: Y) -> dict:
        """
        Generate ``dict`` object representation of context data.

        The representation is intended to be used for tags data in ARM
        template parameter files. Note that "microsoft", "azure" and
        "windows" are reserved prefixes in Azure tags so any fields using
        these prefixes must be renamed for the dictionary conversion.
        """
        return {
            "application_name": self.__application_name,
            "client": self.__client,
            "commit_sha": self.commit_sha,
            "frame_name": self.__frame_name,
            "pipeline_id": self.pipeline_id,
            "release_id": self.release_id,
            "release_state": self.release_state,
            "resource_suffix": self.resource_suffix,
            "subscription_name": self.__azure_subscription_name,
            "system": self.__system,
            "tenant_name": self.__azure_tenant_name,
        }

    def __str__(self: Y) -> str:
        """Convert object to str for logging purposes."""
        return str(self.as_dict())


X = typing.TypeVar("X", bound="DeployDataView")


class DeployDataView:
    """Data critical to a resource deployment."""

    azure_credentials: AzureCredentials
    deployment_tuple: str
    location_primary: str
    release_state: str
    static_secrets: dict

    frame_folder: typing.Optional[pathlib.Path] = None

    __location_secondary: typing.Optional[str] = None

    iteration_context: IterationContext
    to: StructuredTo
    __puff_map: typing.Optional[PuffMap] = None

    def __init__(
        self: X,
        azure_credentials: AzureCredentials,
        deployment_tuple: str,
        location_primary: str,
        release_state: str,
        static_secrets: dict,
        location_secondary: typing.Optional[str] = None,
    ) -> None:
        """Construct ``DeployDataView`` object."""
        self.azure_credentials = azure_credentials
        self.deployment_tuple = deployment_tuple
        self.location_primary = location_primary
        self.release_state = release_state
        self.static_secrets = static_secrets
        self.__location_secondary = location_secondary

        self.iteration_context = IterationContext()
        self.to = StructuredTo()

    @property
    def location_secondary(self: X) -> typing.Optional[str]:
        """Get location secondary property."""
        return self.__location_secondary

    @location_secondary.setter
    def location_secondary(self: X, v: str) -> None:
        """Set location secondary property."""
        self.__location_secondary = v

    @property
    def puff_map(self: X) -> PuffMap:
        """Get puff map property."""
        if self.__puff_map:
            return self.__puff_map
        else:
            raise PipelineViewError("puff map not specified in deployment data")

    @puff_map.setter
    def puff_map(self: X, value: PuffMap) -> None:
        """Set puff map property."""
        self.__puff_map = value

    def __str__(self: X) -> str:
        """Convert object to str for logging purposes."""
        return str(
            {
                "azure_credentials": self.azure_credentials,
                "deployment_tuple": self.deployment_tuple,
                "location_primary": self.location_primary,
                "release_state": self.release_state,
                "location_secondary": self.__location_secondary,
                "iteration_context": self.iteration_context,
            }
        )


W = typing.TypeVar("W", bound="FlattenedDeployment")


@dataclasses.dataclass
class FlattenedDeployment:
    """Flattened deployment data for deployment concurrency."""

    context: DeploymentContext
    data: DeployDataView

    def copy_add_frame(self: W, frame_name: str) -> W:
        """
        Deep copy this object and add frame name data.

        Args:
            frame_name: Frame name data to add.

        Returns:
            Copied and updated object.
        """
        x = copy.deepcopy(self)
        x.context.frame_name = frame_name
        x.data.iteration_context.append(frame_name)
        return x

    def copy_add_application(self: W, application_name: str) -> W:
        """
        Deep copy this object and add application name data.

        Args:
            application_name: Application name data to add.

        Returns:
            Copied and updated object.
        """
        x = copy.deepcopy(self)
        x.context.application_name = application_name
        x.data.iteration_context.append(application_name)
        return x


V = typing.TypeVar("V", bound="SubscriptionView")
U = typing.TypeVar("U", bound="DeploymentView")
T = typing.TypeVar("T", bound="ReleaseView")


class SubscriptionView:
    """View of pipeline configuration conditioned by a subscription name."""

    deployment_view: "DeploymentView"
    subscription_name: str

    def __init__(self: V, deployment_view: U, subscription_name: str) -> None:
        """Construct ``SubscriptionView`` object."""
        self.deployment_view = deployment_view
        self.subscription_name = subscription_name

        self._validate_subscription()

    @property
    def deploy_data(self: V) -> typing.List[DeployDataView]:
        """Provide deployment data for each location in this subscription."""
        base_data = (
            self.deployment_view.release_view.configuration.subscriptions[
                self.subscription_name
            ]
        )
        result: typing.List[DeployDataView] = list()
        this_deployment = (
            self.deployment_view.release_view.configuration.deployments[
                str(self.deployment_view.deployment_tuple)
            ]
        )
        this_service_principals = (
            self.deployment_view.release_view.configuration.service_principals  # noqa E501
        )
        if not this_service_principals:
            raise PipelineViewError("Missing service principal credentials")

        this_tenants = self.deployment_view.release_view.configuration.tenants
        for this_locations in this_deployment.subscriptions[
            self.subscription_name
        ].locations:
            this_credentials = AzureCredentials(
                userid=this_service_principals[self.subscription_name].id,
                secret=this_service_principals[self.subscription_name].secret,
                name=this_service_principals[self.subscription_name].name,
                subscription=base_data.azure_id,
                tenant=this_tenants[base_data.tenant].azure_id,
            )
            static_secrets = dict()
            if self.deployment_view.release_view.configuration.static_secrets:
                static_secrets = self.deployment_view.release_view.configuration.static_secrets[  # noqa E501
                    self.subscription_name
                ]

            this_data: typing.Dict[str, typing.Any] = {
                "azure_credentials": this_credentials,
                "deployment_tuple": str(self.deployment_view.deployment_tuple),
                "location_primary": this_locations.primary,
                "location_secondary": None,
                "release_state": self.deployment_view.release_view.deployment_context.release_state,  # noqa E501
                "static_secrets": static_secrets,
            }
            if this_locations.secondary:
                this_data["location_secondary"] = this_locations.secondary

            result.append(DeployDataView(**this_data))

        return result

    def _validate_subscription(self: V) -> None:
        if (
            self.subscription_name
            not in self.deployment_view.release_view.configuration.subscriptions
        ):
            raise PipelineViewError(
                "Bad subscription, {0}".format(str(self.subscription_name))
            )


class DeploymentView:
    """View of pipeline configuration conditioned by deployment state."""

    release_view: "ReleaseView"
    deployment_tuple: DeploymentTuple

    def __init__(
        self: U, release_view: T, deployment_state: DeploymentTuple
    ) -> None:
        """Construct ``DeploymentView`` object."""
        self.release_view = release_view
        self.deployment_tuple = deployment_state

        self._validate_deployment_tuple()

    @property
    def subscriptions(self: U) -> typing.List[SubscriptionView]:
        """Provide the subscriptions in this deployment."""
        result: typing.List[SubscriptionView] = list()
        if (
            str(self.deployment_tuple)
            in self.release_view.configuration.deployments
        ):
            for (
                this_subscription
            ) in self.release_view.configuration.deployments[
                str(self.deployment_tuple)
            ].subscriptions.keys():
                result.append(SubscriptionView(self, this_subscription))

        return result

    def _validate_deployment_tuple(self: U) -> None:
        if (
            self.deployment_tuple.release_state
            not in self.release_view.configuration.release_states
        ):
            raise PipelineViewError(
                "Bad release state, {0}".format(str(self.deployment_tuple))
            )

        if (
            self.deployment_tuple.client
            not in self.release_view.configuration.clients.keys()
        ):
            raise PipelineViewError(
                "Bad client, {0}".format(str(self.deployment_tuple))
            )

        if (
            self.deployment_tuple.system
            not in self.release_view.configuration.systems
        ):
            raise PipelineViewError(
                "Bad system, {0}".format(str(self.deployment_tuple))
            )


class ReleaseView:
    """View of pipeline configuration conditioned by release state."""

    configuration: PipelineConfiguration
    deployment_context: DeploymentContext

    def __init__(
        self: T,
        configuration: PipelineConfiguration,
        deployment_context: DeploymentContext,
    ) -> None:
        """Construct ``ReleaseView`` object."""
        self.configuration = configuration
        self.deployment_context = deployment_context

        self._validate_release_state()

    @property
    def deployments(self: T) -> typing.List[DeploymentView]:
        """Provide the deployments in this release."""
        result: typing.List[DeploymentView] = list()
        for this_client in self.configuration.clients.keys():
            for this_system in self.configuration.systems:
                this_state = DeploymentTuple(
                    client=this_client,
                    release_state=self.deployment_context.release_state,
                    system=this_system,
                )
                if str(this_state) in self.configuration.deployments.keys():
                    result.append(DeploymentView(self, this_state))

        return result

    def flatten(
        self: T,
        to: StructuredTo,
    ) -> typing.List[FlattenedDeployment]:
        """Flatten the nested hierarchy of views into a simple list."""
        result: typing.List[FlattenedDeployment] = list()
        for this_deployment in self.deployments:
            for this_subscription in this_deployment.subscriptions:
                for this_deploy_data in this_subscription.deploy_data:
                    updated_context = copy.deepcopy(self.deployment_context)
                    updated_context.azure_subscription_name = (
                        this_subscription.subscription_name
                    )
                    updated_context.azure_tenant_name = (
                        self.configuration.subscriptions[
                            this_subscription.subscription_name
                        ].tenant
                    )
                    updated_context.client = (
                        this_deployment.deployment_tuple.client
                    )
                    updated_context.system = (
                        this_deployment.deployment_tuple.system
                    )

                    updated_data = copy.deepcopy(this_deploy_data)
                    updated_data.to = to

                    this_value = FlattenedDeployment(
                        context=updated_context,
                        data=updated_data,
                    )
                    result.append(this_value)
        return result

    def _validate_release_state(self: T) -> None:
        if (
            self.deployment_context.release_state
            not in self.configuration.release_states
        ):
            message = "Bad release state, {0}, {1}".format(
                self.deployment_context.release_state,
                str(self.configuration.release_states),
            )
            raise PipelineViewError(message)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import dataclasses
import enum


@dataclasses.dataclass
class PipelineCliOptions:
    """Pipeline configuration options provided from CLI."""

    enable_validation: bool
    monitor_sleep_seconds: float
    wait_timeout_seconds: float


@enum.unique
class ExitState(enum.Enum):
    """Exit state definitions for ``deploy-me`` utility."""

    UNKNOWN_ERROR = 100
    BAD_CLI_ARGUMENTS = 101
    BAD_DEPLOYMENT_CONFIGURATION = 102
    DEPLOYMENT_FAILED = 103
    DEPLOYMENT_TIMEOUT = 104
    DEPLOYMENT_CANCELLED = 105



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import os
import pathlib
import tempfile
import typing


@contextlib.contextmanager
def simple_isolated_filesystem() -> typing.Generator[pathlib.Path, None, None]:
    with tempfile.TemporaryDirectory() as dir:
        d = pathlib.Path(dir)
        yield d



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import logging

import pytest

from foodx_devops_tools.pipeline_config import (
    FramesDefinition,
    PipelineConfiguration,
    PuffMapGeneratedFiles,
)
from foodx_devops_tools.pipeline_config.exceptions import (
    PipelineConfigurationError,
)
from tests.ci.support.pipeline_config import MOCK_PATHS, MOCK_SECRET

log = logging.getLogger(__name__)


def test_mismatched_frames_raises1(mock_loads, mock_results):
    mock_results.frames = FramesDefinition.parse_obj(
        {
            "frames": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": [
                                {
                                    "resource_group": "a1_group",
                                    "name": "a1stp1",
                                    "mode": "Incremental",
                                },
                            ],
                        },
                        "folder": "some/path",
                    },
                },
            },
        }
    ).frames
    mock_results.puff_map = PuffMapGeneratedFiles.parse_obj(
        {
            "puff_map": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": {
                                "arm_parameters_files": {
                                    "r1": {
                                        "sub1": {
                                            "a1stp1": "some/path/puff1.json"
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "f2": {
                        "applications": {
                            "a2": {
                                "arm_parameters_files": {
                                    "r1": {
                                        "sub1": {
                                            "a2stp1": "some/path/puff1.json"
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
            },
        }
    ).puff_map
    mock_loads(mock_results)

    with pytest.raises(
        PipelineConfigurationError,
        match=r"Frame definitions mismatch between frames and puff map",
    ):
        PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)


def test_mismatched_frames_raises2(mock_loads, mock_results):
    mock_results.frames = FramesDefinition.parse_obj(
        {
            "frames": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": [
                                {
                                    "resource_group": "a1_group",
                                    "name": "a1stp1",
                                    "mode": "Incremental",
                                },
                            ],
                        },
                        "folder": "some/path",
                    },
                    "f2": {
                        "applications": {
                            "a2": [
                                {
                                    "resource_group": "a2_group",
                                    "name": "a2stp1",
                                    "mode": "Incremental",
                                },
                            ],
                        },
                        "folder": "some/path",
                    },
                },
            },
        }
    ).frames
    mock_results.puff_map = PuffMapGeneratedFiles.parse_obj(
        {
            "puff_map": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": {
                                "arm_parameters_files": {
                                    "r1": {
                                        "sub1": {
                                            "a1stp1": "some/path/puff1.json"
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
            },
        }
    ).puff_map
    mock_loads(mock_results)

    with pytest.raises(
        PipelineConfigurationError,
        match=r"Frame definitions mismatch between frames and puff map",
    ):
        PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)


def test_mismatched_applications_raises1(mock_loads, mock_results):
    mock_results.frames = FramesDefinition.parse_obj(
        {
            "frames": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": [
                                {
                                    "resource_group": "a1_group",
                                    "name": "a1stp1",
                                    "mode": "Incremental",
                                },
                            ],
                        },
                        "folder": "some/path",
                    },
                },
            },
        }
    ).frames
    mock_results.puff_map = PuffMapGeneratedFiles.parse_obj(
        {
            "puff_map": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": {
                                "arm_parameters_files": {
                                    "r1": {
                                        "sub1": {
                                            "a1stp1": "some/path/puff1.json"
                                        }
                                    }
                                }
                            },
                            "a2": {
                                "arm_parameters_files": {
                                    "r1": {
                                        "sub1": {
                                            "a2stp1": "some/path/puff1.json"
                                        }
                                    }
                                }
                            },
                        }
                    },
                },
            },
        }
    ).puff_map
    mock_loads(mock_results)

    with pytest.raises(
        PipelineConfigurationError,
        match=r"Application definitions mismatch between frames and puff "
        r"map",
    ):
        PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)


def test_mismatched_applications_raises2(mock_loads, mock_results):
    mock_results.frames = FramesDefinition.parse_obj(
        {
            "frames": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": [
                                {
                                    "resource_group": "a1_group",
                                    "name": "a1stp1",
                                    "mode": "Incremental",
                                },
                            ],
                            "a2": [
                                {
                                    "resource_group": "a2_group",
                                    "name": "a2stp1",
                                    "mode": "Incremental",
                                },
                            ],
                        },
                        "folder": "some/path",
                    },
                },
            },
        }
    ).frames
    mock_results.puff_map = PuffMapGeneratedFiles.parse_obj(
        {
            "puff_map": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": {
                                "arm_parameters_files": {
                                    "r1": {
                                        "sub1": {
                                            "a1stp1": "some/path/puff1.json"
                                        }
                                    }
                                }
                            },
                        }
                    },
                },
            },
        }
    ).puff_map
    mock_loads(mock_results)

    with pytest.raises(
        PipelineConfigurationError,
        match=r"Application definitions mismatch between frames and puff "
        r"map",
    ):
        PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)


def test_bad_puff_map_release_state_raises(mock_loads, mock_results):
    mock_results.frames = FramesDefinition.parse_obj(
        {
            "frames": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": [
                                {
                                    "resource_group": "a1_group",
                                    "name": "a1stp1",
                                    "mode": "Incremental",
                                },
                            ],
                        },
                        "folder": "some/path",
                    },
                },
            },
        }
    ).frames
    mock_results.puff_map = PuffMapGeneratedFiles.parse_obj(
        {
            "puff_map": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": {
                                "arm_parameters_files": {
                                    "bad_state": {
                                        "sub1": {
                                            "a1stp1": "some/path/puff1.json"
                                        }
                                    }
                                }
                            },
                        }
                    },
                },
            },
        }
    ).puff_map
    mock_loads(mock_results)

    with pytest.raises(
        PipelineConfigurationError, match=r"Bad release state in puff map"
    ):
        PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)


def test_bad_puff_map_subscription_raises(mock_loads, mock_results):
    mock_results.frames = FramesDefinition.parse_obj(
        {
            "frames": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": [
                                {
                                    "resource_group": "a1_group",
                                    "name": "a1stp1",
                                    "mode": "Incremental",
                                },
                            ],
                        },
                        "folder": "some/path",
                    },
                },
            },
        }
    ).frames
    mock_results.puff_map = PuffMapGeneratedFiles.parse_obj(
        {
            "puff_map": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": {
                                "arm_parameters_files": {
                                    "r1": {
                                        "bad_sub": {
                                            "a1stp1": "some/path/puff1.json"
                                        }
                                    }
                                }
                            },
                        }
                    },
                },
            },
        }
    ).puff_map
    mock_loads(mock_results)

    with pytest.raises(
        PipelineConfigurationError, match=r"Bad subscription in puff map"
    ):
        PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)


def test_missing_application_step_raises1(mock_loads, mock_results):
    mock_results.frames = FramesDefinition.parse_obj(
        {
            "frames": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": [
                                {
                                    "resource_group": "a1_group",
                                    "name": "a1stp1",
                                    "mode": "Incremental",
                                },
                                {
                                    "resource_group": "a1_group",
                                    "name": "a1stp2",
                                    "mode": "Incremental",
                                },
                            ],
                        },
                        "folder": "some/path",
                    },
                },
            },
        }
    ).frames
    mock_results.puff_map = PuffMapGeneratedFiles.parse_obj(
        {
            "puff_map": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": {
                                "arm_parameters_files": {
                                    "r1": {
                                        "sys1_c1_r1a": {
                                            "a1stp2": "some/path/puff1.json"
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            },
        }
    ).puff_map
    mock_loads(mock_results)

    with pytest.raises(
        PipelineConfigurationError,
        match=r"Application step name mismatch between frames and puff map",
    ):
        PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)


def test_missing_application_step_raises2(mock_loads, mock_results):
    mock_results.frames = FramesDefinition.parse_obj(
        {
            "frames": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": [
                                {
                                    "resource_group": "a1_group",
                                    "name": "a1stp2",
                                    "mode": "Incremental",
                                },
                            ],
                        },
                        "folder": "some/path",
                    },
                },
            },
        }
    ).frames
    mock_results.puff_map = PuffMapGeneratedFiles.parse_obj(
        {
            "puff_map": {
                "frames": {
                    "f1": {
                        "applications": {
                            "a1": {
                                "arm_parameters_files": {
                                    "r1": {
                                        "sys1_c1_r1a": {
                                            "a1stp1": "some/path/puff1.json",
                                            "a1stp2": "some/path/puff2.json",
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            },
        }
    ).puff_map
    mock_loads(mock_results)

    with pytest.raises(
        PipelineConfigurationError,
        match=r"Application step name mismatch between frames and puff map",
    ):
        PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)



# https://gitlab.com/ci-cd-devops/build_harness/-/blob/feature/27-fix-main-test-publish/build_harness/commands/_logging.py
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.

#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Logging configuration."""

import logging
import logging.handlers
import pathlib
import typing

import pydantic

from ._declarations import (
    DEFAULT_FILE_ROTATION_BACKUPS,
    DEFAULT_FILE_ROTATION_ENABLED,
    DEFAULT_FILE_ROTATION_SIZE_MB,
)

log = logging.getLogger(__name__)

T = typing.TypeVar("T", bound="LoggingState")


class LoggingConfiguration(pydantic.BaseModel):
    """Logging configuration data."""

    disable_file_logging: bool
    enable_console_logging: bool
    enable_file_rotation: bool
    file_rotation_size_megabytes: int
    log_file_path: typing.Optional[pathlib.Path]
    log_level: int
    max_rotation_backup_files: int


class LoggingState:
    """Logging configuration parameters."""

    configuration: LoggingConfiguration
    rotation_handler: typing.Optional[logging.handlers.RotatingFileHandler]

    def __init__(
        self: T,
        disable_file_logging: bool,
        enable_console_logging: bool,
        log_level_text: str,
        default_log_file: pathlib.Path,
    ) -> None:
        """
        Construct ``LoggingState`` object.

        Args:
            disable_file_logging: Disable file logging flag.
            enable_console_logging: Enable console logging flag.
        """
        self.configuration = LoggingConfiguration(
            **{
                "disable_file_logging": disable_file_logging,
                "enable_console_logging": enable_console_logging,
                "enable_file_rotation": DEFAULT_FILE_ROTATION_ENABLED,
                "file_rotation_size_megabytes": DEFAULT_FILE_ROTATION_SIZE_MB,
                "log_file_path": (
                    default_log_file if not disable_file_logging else None
                ),
                "log_level": getattr(logging, log_level_text.upper()),
                "max_rotation_backup_files": DEFAULT_FILE_ROTATION_BACKUPS,
            }
        )
        self.rotation_handler = None

        self.set_logging_state()

    def set_logging_state(self: T) -> None:
        """Apply the logging state from configured parameters."""
        root_logger = logging.getLogger()
        self.__set_log_level(root_logger)
        self.__set_file_logging(root_logger)
        self.__set_console_logging(root_logger)

        log.info("Logging state set")
        log.debug(
            "Applied logging configuration, {0}".format(
                self.configuration.dict()
            )
        )

    def __set_file_logging(self: T, root_logger: logging.Logger) -> None:
        """
        Enable or disable file logging.

        Args:
            root_logger: Root logger to modify.
        """
        if not self.configuration.disable_file_logging:
            # No change if a rotation handler already exists.
            if not self.rotation_handler:
                this_handler = logging.handlers.RotatingFileHandler(
                    str(self.configuration.log_file_path),
                    backupCount=self.configuration.max_rotation_backup_files,
                    maxBytes=self.configuration.file_rotation_size_megabytes
                    * (1024 ** 2),
                )
                this_handler.setLevel(self.configuration.log_level)
                self.rotation_handler = this_handler

                root_logger.addHandler(self.rotation_handler)
        elif self.rotation_handler:
            root_logger.removeHandler(self.rotation_handler)
            self.rotation_handler = None
        # else self.rotation_handler is None and self.disable_file_logging
        # so do nothing

    def __set_console_logging(self: T, root_logger: logging.Logger) -> None:
        """
        Enable or disable console logging.

        Args:
            root_logger: Root logger to modify.
        """

        def remove_stream_handlers() -> None:
            for this_handler in root_logger.handlers:
                if type(this_handler) == logging.StreamHandler:
                    root_logger.removeHandler(this_handler)

        remove_stream_handlers()
        if self.configuration.enable_console_logging:
            console_handler = logging.StreamHandler()
            console_handler.setLevel(self.configuration.log_level)

            root_logger.addHandler(console_handler)

    def __set_log_level(self: T, root_logger: logging.Logger) -> None:
        """
        Set log level on any existing handlers.

        Args:
            root_logger: Root logger to modify.
        """
        for this_handler in root_logger.handlers:
            this_handler.setLevel(self.configuration.log_level)
        # Ensure that the logging level propagates to any subsequently
        # created handlers.
        root_logger.setLevel(self.configuration.log_level)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Static secrets deployment configuration I/O."""

import logging
import pathlib
import typing

import pydantic

from foodx_devops_tools.utilities.ansible import AnsibleVaultError
from foodx_devops_tools.utilities.io import load_encrypted_data

from ._exceptions import StaticSecretsError

log = logging.getLogger(__name__)


ValueType = typing.Dict[str, typing.Dict[str, typing.Any]]


class StaticSecrets(pydantic.BaseModel):
    """Define a collection of static secrets."""

    static_secrets: ValueType


def load_static_secrets(
    secrets_paths: typing.Set[pathlib.Path], decrypt_token: str
) -> StaticSecrets:
    """Load static secrets from the relevant directory and it's files."""
    try:
        secrets_data: dict = {
            "static_secrets": dict(),
        }
        for this_path in secrets_paths:
            if this_path.is_file():
                yaml_data = load_encrypted_data(this_path, decrypt_token)
                if "static_secrets" in yaml_data:
                    secrets_data["static_secrets"].update(
                        yaml_data["static_secrets"]
                    )
                else:
                    log.error(
                        f"static_secrets field not present in file, {this_path}"
                    )

        this_object = StaticSecrets.parse_obj(secrets_data)

        return this_object
    except (AnsibleVaultError, NotADirectoryError, FileNotFoundError) as e:
        raise StaticSecretsError(str(e)) from e



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib
import tempfile

from foodx_devops_tools.utilities.io import acquire_token, load_encrypted_data
from tests.ci.support.ansible import encrypted_file


def test_acquire_token():
    with tempfile.TemporaryDirectory() as dir:
        this_dir = pathlib.Path(dir)
        this_file = this_dir / "some.file"
        with this_file.open(mode="w") as f:
            f.write("something")

        with this_file.open("r") as f:
            result = acquire_token(f)

        assert result == "something"


def test_load_encrypted_data():
    file_content = """---
some:
  data: value
"""
    decrypt_token = "somesecret"
    with encrypted_file(file_content, decrypt_token) as (encrypted_file_path):

        result = load_encrypted_data(encrypted_file_path, decrypt_token)

        assert result == {"some": {"data": "value"}}



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Manage templating of deployment definitions."""

import pathlib
import re

import aiofiles  # type: ignore

ESCOLAR_REGEX = r"(?P<reference>#(?P<file_path>.+?)#)"


async def _do_substitutions(source_content: str) -> str:
    escolards = re.finditer(ESCOLAR_REGEX, source_content)

    modified_content = source_content
    for occurrence in escolards:
        snippet_file = pathlib.Path(occurrence.group("file_path"))
        async with aiofiles.open(snippet_file, mode="r") as f:
            snippet_content = await f.read()
            # escape double quotes and retain newlines as escapes to embody
            # the snippet in a single line
            escaped_newlines = (
                repr(snippet_content).strip("'").replace(r"\n", r"\\n")
            )
            escaped_quotes = escaped_newlines.replace('"', r'\\"')
            modified_content = re.sub(
                occurrence.group("reference"), escaped_quotes, modified_content
            )

    return modified_content


async def do_snippet_substitution(
    source_file: pathlib.Path, destination_file: pathlib.Path
) -> None:
    """
    Apply template parameters.

    Presently uses the legacy "escolar" way of doing template substitution.
    """
    async with aiofiles.open(source_file, mode="r") as source:
        source_content = await source.read()

    modified_content = await _do_substitutions(source_content)

    async with aiofiles.open(destination_file, mode="w") as target:
        await target.writelines(modified_content)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Declarations of constants for pipeline reference parsing."""

import dataclasses
import re
import typing

BRANCH_PREFIX = r"refs/heads/"
TAG_PREFIX = r"refs/tags/"

SEMANTIC_VERSION_GITREF = r"(\d+\.\d+\.\d+)"

SUBSCRIPTION_NAME = (
    r"(?P<system>[a-z0-9]+)"
    r"_"
    r"(?P<client>[a-z0-9]+)"
    r"_"
    r"(?P<resource_suffix>[a-z0-9]+)"
)


class SubscriptionNameError(Exception):
    """Problem parsing a subscription name into subscription data."""


T = typing.TypeVar("T", bound="SubscriptionData")


@dataclasses.dataclass
class SubscriptionData:
    """Manage components of a subscription name."""

    client: str
    resource_suffix: str
    system: str

    def __str__(self: T) -> str:
        """Convert subscription data to a string."""
        return "{0}_{1}_{2}".format(
            self.system, self.client, self.resource_suffix
        )

    @classmethod
    def from_subscription_name(cls: typing.Type[T], value: str) -> T:
        """Extract subscription data from a string."""
        result = re.match(SUBSCRIPTION_NAME, value)
        if not result:
            raise SubscriptionNameError(
                "Subscription name does not conform to naming pattern "
                "<system>_<client>_<resource_suffix>, {0}".format(value)
            )

        this_object = cls(
            client=result.group("client"),
            resource_suffix=result.group("resource_suffix"),
            system=result.group("system"),
        )

        return this_object



# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from foodx_devops_tools.release_flow_entry import release_flow
from tests.ci.support.click_runner import click_runner  # noqa: F401


class TestAzureSubcommand:
    def test_azure_release_state(self, click_runner):
        mock_input = ["azure", "state", "refs/heads/feature/some/path"]

        result = click_runner.invoke(release_flow, mock_input)

        assert result.exit_code == 0
        assert result.output == "ftr"

    def test_azure_release_id(self, click_runner):
        mock_input = ["azure", "id", "refs/tags/3.14.159-alpha.13"]

        result = click_runner.invoke(release_flow, mock_input)

        assert result.exit_code == 0
        assert result.output == "3.14.159-alpha.13"



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Pydantic model of puff parameter data."""

import typing

import pydantic

PuffRegionVariables = typing.Dict[str, typing.Any]

PuffRegionValues = typing.Dict[str, typing.Optional[PuffRegionVariables]]

T = typing.TypeVar("T", bound="PuffRegion")


class PuffRegion(pydantic.BaseModel):
    """
    Puff parameter region (singular) data model.

    Region data is a dictionary of variable names and values to be applied.
    """

    __root__: PuffRegionValues

    def __getitem__(self: T, item: str) -> typing.Optional[PuffRegionVariables]:
        """Get region parameters."""
        return self.__root__[item]

    def __contains__(self: T, item: str) -> bool:
        """Verify the region contains a parameter."""
        return item in self.__root__

    @pydantic.root_validator(pre=True)
    def check_single_entry(
        cls: pydantic.BaseModel,
        values: typing.Dict[str, typing.Dict[str, typing.Any]],
    ) -> dict:
        """Check that a single name entry exists."""
        assert values
        # ignore this mypy error:
        #   error: Argument 1 to "len" has incompatible type \
        #     "Optional[Dict[str, Any]]"; expected "Sized"  [arg-type]
        if len(values.get("__root__")) > 1:  # type: ignore
            raise ValueError(
                "must only specify a single name in a region entry"
            )
        return values


PuffRegionsValues = typing.List[PuffRegion]


U = typing.TypeVar("U", bound="PuffRegions")


class PuffRegions(pydantic.BaseModel):
    """Puff parameter regions data model."""

    __root__: PuffRegionsValues

    def __getitem__(self: U, item: int) -> PuffRegion:
        """Get a region item."""
        return self.__root__[item]

    @pydantic.root_validator(pre=True)
    def check_regions(cls: pydantic.BaseModel, values: dict) -> dict:
        """
        Check that regions are formulated correctly.

        Args:
            values:

        Returns:
            Reviewed region values.
        """
        this_values = values.get("__root__")
        if not isinstance(this_values, list):
            raise ValueError("")
        for k in range(0, len(this_values)):
            PuffRegion.parse_obj(this_values[k])
        return values


PuffEnvironmentValues = typing.Dict[str, typing.Any]


V = typing.TypeVar("V", bound="PuffEnvironment")


class PuffEnvironment(pydantic.BaseModel):
    """Puff parameter environment (singular) data model."""

    __root__: PuffEnvironmentValues

    @pydantic.root_validator(pre=True)
    def check_regions(cls: pydantic.BaseModel, values: dict) -> dict:
        """Verify validity of region entries."""
        # ignore this mypy error:
        #    error: Unsupported right operand type for in \
        #      ("Optional[Any]")  [operator]
        if "regions" in values.get("__root__"):  # type: ignore
            PuffRegions.parse_obj(values["__root__"]["regions"])
        return values

    def __getitem__(self: V, item: str) -> typing.Any:
        """Get environment item."""
        return self.__root__[item]

    def __contains__(self: V, item: str) -> bool:
        """Verify that an environment contains the specific item."""
        return item in self.__root__


W = typing.TypeVar("W", bound="PuffEnvironments")


class PuffEnvironments(pydantic.BaseModel):
    """Puff parameter environments data model."""

    __root__: typing.Dict[str, typing.Optional[PuffEnvironment]]

    def __getitem__(self: W, item: str) -> typing.Optional[PuffEnvironment]:
        """Get named environment entry."""
        return self.__root__[item]

    def __contains__(self: W, item: str) -> bool:
        """Verify that environments contains the named environment."""
        return item in self.__root__


PuffServiceValues = typing.Dict[str, typing.Any]


X = typing.TypeVar("X", bound="PuffService")


class PuffService(pydantic.BaseModel):
    """Puff parameter service (singular) data model."""

    __root__: PuffServiceValues

    @pydantic.root_validator(pre=True)
    def check_environments(cls: pydantic.BaseModel, values: dict) -> dict:
        """Check validity of environments structure."""
        if "environments" in values:
            PuffEnvironments.parse_obj(values["environments"])
        return values

    def __getitem__(self: X, item: str) -> typing.Any:
        """Get the specified item from a service."""
        return self.__root__[item]

    def __contains__(self: X, item: str) -> bool:
        """Verify that a service contains the named item."""
        return item in self.__root__


Y = typing.TypeVar("Y", bound="PuffServices")


class PuffServices(pydantic.BaseModel):
    """Puff parameter services data model."""

    __root__: typing.Dict[str, PuffService]

    def __getitem__(self: Y, item: str) -> PuffService:
        """Get the named service."""
        return self.__root__[item]

    def __contains__(self: Y, item: str) -> bool:
        """Verify that the named service is in the services list."""
        return item in self.__root__


PuffModelValues = typing.Dict[str, typing.Any]


Z = typing.TypeVar("Z", bound="PuffParameterModel")


class PuffParameterModel(pydantic.BaseModel):
    """Data model for puff parameters."""

    __root__: PuffModelValues

    def __getitem__(self: Z, item: str) -> typing.Any:
        """Get the name item from the puff parameters."""
        return self.__root__[item]

    def __contains__(self: Z, item: str) -> bool:
        """Verify that the puff parameters contains the specified item."""
        return item in self.__root__

    @pydantic.root_validator(pre=True)
    def check_structured_elements(
        cls: pydantic.BaseModel, values: dict
    ) -> dict:
        """Check validity of puff parameters structure."""
        if "environments" in values["__root__"]:
            PuffEnvironments.parse_obj(values["__root__"]["environments"])

        if "services" in values["__root__"]:
            PuffServices.parse_obj(values["__root__"]["services"])

        return values



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import json
import pathlib
import tempfile

import pytest

from foodx_devops_tools.templates import do_snippet_substitution


async def do_check(source_file, target_file, expected_data):
    await do_snippet_substitution(source_file, target_file)

    with target_file.open(mode="r") as f:
        content = f.read()
        read_data = json.loads(content)

    assert read_data == expected_data


@pytest.mark.asyncio
async def test_no_change(mock_async_method):
    with tempfile.TemporaryDirectory() as dir:
        this_dir = pathlib.Path(dir)

        data = {
            "something": "some content",
        }
        content = json.dumps(data)

        source_file = this_dir / "this_file"
        with source_file.open(mode="w") as f:
            f.write(content)

        target_file = this_dir / "that_file"
        expected_data = {
            "something": "some content",
        }
        await do_check(source_file, target_file, expected_data)


@pytest.mark.asyncio
async def test_single(mock_async_method):
    snippet_content = "snippet content"
    with tempfile.TemporaryDirectory() as dir:
        this_dir = pathlib.Path(dir)

        snippet_file = this_dir / "this_snippet"
        data = {
            "something": "#{0}#".format(snippet_file),
        }
        content = json.dumps(data)

        source_file = this_dir / "this_file"
        with source_file.open(mode="w") as f:
            f.write(content)

        with snippet_file.open(mode="w") as f:
            f.write(snippet_content)

        target_file = this_dir / "that_file"
        expected_data = {
            "something": "snippet content",
        }
        await do_check(source_file, target_file, expected_data)


@pytest.mark.asyncio
async def test_multiline_snippet(mock_async_method):
    snippet_content = """snippet content
on multiple
lines
"""
    with tempfile.TemporaryDirectory() as dir:
        this_dir = pathlib.Path(dir)

        snippet_file = this_dir / "this_snippet"
        data = {
            "something": "#{0}#".format(snippet_file),
        }
        content = json.dumps(data)

        source_file = this_dir / "this_file"
        with source_file.open(mode="w") as f:
            f.write(content)

        with snippet_file.open(mode="w") as f:
            f.write(snippet_content)

        target_file = this_dir / "that_file"
        expected_data = {
            "something": "snippet content\non multiple\nlines\n",
        }
        await do_check(source_file, target_file, expected_data)


@pytest.mark.asyncio
async def test_quoted_snippet(mock_async_method):
    snippet_content = """snippet content="is quoted" lines"""
    with tempfile.TemporaryDirectory() as dir:
        this_dir = pathlib.Path(dir)

        snippet_file = this_dir / "this_snippet"
        data = {
            "something": "#{0}#".format(snippet_file),
        }
        content = json.dumps(data)

        source_file = this_dir / "this_file"
        with source_file.open(mode="w") as f:
            f.write(content)

        with snippet_file.open(mode="w") as f:
            f.write(snippet_content)

        target_file = this_dir / "that_file"
        expected_data = {
            "something": 'snippet content="is quoted" lines',
        }
        await do_check(source_file, target_file, expected_data)


@pytest.mark.asyncio
async def test_quoted_multiline_snippet(mock_async_method):
    snippet_content = """snippet content
"is quoted"
lines
"""
    with tempfile.TemporaryDirectory() as dir:
        this_dir = pathlib.Path(dir)

        snippet_file = this_dir / "this_snippet"
        data = {
            "something": "#{0}#".format(snippet_file),
        }
        content = json.dumps(data)

        source_file = this_dir / "this_file"
        with source_file.open(mode="w") as f:
            f.write(content)

        with snippet_file.open(mode="w") as f:
            f.write(snippet_content)

        target_file = this_dir / "that_file"
        expected_data = {
            "something": 'snippet content\n"is quoted"\nlines\n',
        }
        await do_check(source_file, target_file, expected_data)


@pytest.mark.asyncio
async def test_multiple_lines(mock_async_method):
    snippet1_content = "snippet1 content"
    snippet2_content = "more content"
    with tempfile.TemporaryDirectory() as dir:
        this_dir = pathlib.Path(dir)

        snippet1 = this_dir / "snippet1"
        snippet2 = this_dir / "snippet2"
        data = {
            "one": "#{0}#".format(snippet1),
            "none": "no change",
            "two": "#{0}#".format(snippet2),
        }
        content = json.dumps(data)

        source_file = this_dir / "this_file"
        with source_file.open(mode="w") as f:
            f.write(content)

        with snippet1.open(mode="w") as f:
            f.write(snippet1_content)
        with snippet2.open(mode="w") as f:
            f.write(snippet2_content)

        target_file = this_dir / "that_file"
        expected_data = {
            "one": "snippet1 content",
            "none": "no change",
            "two": "more content",
        }
        await do_check(source_file, target_file, expected_data)


@pytest.mark.asyncio
async def test_multiple_in_line(mock_async_method):
    snippet1_content = "snippet1 content"
    snippet2_content = "more content"
    with tempfile.TemporaryDirectory() as dir:
        this_dir = pathlib.Path(dir)

        snippet1 = this_dir / "snippet1"
        snippet2 = this_dir / "snippet2"
        data = {
            "one": "#{0}#joiner#{1}#".format(snippet1, snippet2),
            "none": "no change",
        }
        content = json.dumps(data)

        source_file = this_dir / "this_file"
        with source_file.open(mode="w") as f:
            f.write(content)

        with snippet1.open(mode="w") as f:
            f.write(snippet1_content)
        with snippet2.open(mode="w") as f:
            f.write(snippet2_content)

        target_file = this_dir / "that_file"
        expected_data = {
            "one": "snippet1 contentjoinermore content",
            "none": "no change",
        }
        await do_check(source_file, target_file, expected_data)
